from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, BinaryIO, Final

from dissect.etl.etl import ETL, Event
from flow.record.base import Record

from dissect.target.exceptions import FilesystemError, UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.datetime import parse_tzi

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from flow.record import Record

    from dissect.target.target import Target


class EtlRecordBuilder:
    RECORD_NAME = "filesystem/windows/etl"

    def __init__(self):
        self._create_event_descriptor = lru_cache(4096)(self._create_event_descriptor)

    def _build_record(self, etl_event: Event, etl_path: Path, target: Target) -> Record:
        """Builds an ETL event record."""

        record_values = {}
        record_fields = [
            ("datetime", "ts"),
            ("path", "path"),
            ("string", "ProviderName"),
            ("string", "ProviderId"),
            ("string", "EventType"),
        ]
        record_values["ts"] = etl_event.ts()
        record_values["path"] = etl_path
        record_values["ProviderName"] = etl_event.provider_name()
        record_values["ProviderId"] = etl_event.provider_id()
        record_values["EventType"] = etl_event.symbol()
        record_values["_target"] = target

        for key, value in etl_event.event_values().items():
            record_type = "bytes"
            if key == "TimeZoneInformation":
                # Pretty print TimezoneInformation
                value = parse_tzi(bytes(value))
                record_type = "string"
            elif isinstance(value, list):
                record_type = "string[]"
            elif isinstance(value, int):
                record_type = "varint"
            elif isinstance(value, str):
                record_type = "string"

            record_fields.append((record_type, key))
            record_values[key] = value

        # tuple conversion here is needed for lru_cache
        desc = self._create_event_descriptor(tuple(record_fields))
        return desc(**record_values)

    def _create_event_descriptor(self, record_fields: list[tuple[str, str]]) -> TargetRecordDescriptor:
        return TargetRecordDescriptor(self.RECORD_NAME, record_fields)

    def read_etl_records(self, fh: BinaryIO, path: Path, target: Target) -> Iterator[Record]:
        etl_file = ETL(fh)
        for buffer in etl_file.buffers():
            for event_records in buffer:
                yield self._build_record(event_records.event, path, target)


class EtlPlugin(Plugin):
    """Plugin for parsing Windows ETL Files (``*.etl``)."""

    __namespace__ = "etl"

    PATHS: Final[dict[str, list[str]]] = {
        "boot": [
            "%windir%/system32/wdi/logfiles/bootckcl.etl",
            "%windir%/system32/wdi/logfiles/bootperfdiaglogger.etl",
        ],
        "shutdown": [
            "%windir%/system32/wdi/logfiles/shutdownckcl.etl",
            "%windir%/system32/wdi/logfiles/shutdownperfdiaglogger.etl",
        ],
    }

    def __init__(self, target: Target):
        super().__init__(target)
        self._etl_record_builder = EtlRecordBuilder()

    def check_compatible(self) -> None:
        etl_paths = (etl_file for etl_paths in self.PATHS.values() for etl_file in etl_paths)
        plugin_target_folders = [self.target.resolve(file).exists() for file in etl_paths]
        if not any(plugin_target_folders):
            raise UnsupportedPluginError("No ETL paths found")

    def read_etl_files(self, etl_paths: list[str]) -> Iterator[Record]:
        """Read ETL files using an EtlReader."""
        for etl_path in etl_paths:
            entry = self.target.resolve(etl_path)
            if not entry.exists():
                self.target.log.warning("ETL file does not exist: %s", entry)
                continue

            try:
                entry_data = entry.open()
            except FilesystemError:
                self.target.log.exception("Failed to open ETL file: %s", entry)
                continue

            etl_records = self._etl_record_builder.read_etl_records(entry_data, entry, self.target)
            yield from etl_records

    @export(record=DynamicDescriptor(["datetime"]))
    def etl(self) -> Iterator[DynamicDescriptor]:
        """Return the contents of the ETL files generated at last boot and last shutdown.

        An event trace log (.etl) file, also known as a trace log, stores the trace messages generated during one or
        more trace sessions. A trace session is period in which a trace provider (a component of a user-mode
        application or kernel-mode driver that uses Event Tracing for Windows (ETW) technology to generate trace
        messages or trace events) is generating trace messages.

        References:
            - https://www.hecfblog.com/2018/06/etw-event-tracing-for-windows-and-etl.html
            - https://docs.microsoft.com/en-us/windows-hardware/drivers/devtest/trace-log

        Yields dynamically created records based on the fields inside an ETL event.
        At least contains the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The TimeCreated_SystemTime field of the event.
            Provider_Name (string): The Provider_Name field of the event.
            EventType (string): The type of the event defined by the manifest file.
        """
        for etl_plugin in self.PATHS:
            yield from getattr(self, etl_plugin)()

    @export(record=DynamicDescriptor(["datetime"]))
    def shutdown(self) -> Iterator[DynamicDescriptor]:
        """Return the contents of the ETL files created at last shutdown.

        The plugin reads the content from the ShutdownCKCL.etl file or the ShutdownPerfDiagLogger.etl file (depending
        on the Windows version).

        Yields dynamically created records based on the fields inside an ETL event.
        At least contains the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The TimeCreated_SystemTime field of the event.
            Provider_Name (string): The Provider_Name field of the event.
            EventType (string): The type of the event defined by the manifest file.
        """
        yield from self.read_etl_files(self.PATHS["shutdown"])

    @export(record=DynamicDescriptor(["datetime"]))
    def boot(self) -> Iterator[DynamicDescriptor]:
        """Return the contents of the ETL files created at last boot.

        The plugin reads the content from the BootCKCL.etl file or the BootPerfDiagLogger.etl file (depending
        on the Windows version).

        Yields dynamically created records based on the fields inside an ETL event.
        At least contains the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The TimeCreated_SystemTime field of the event.
            Provider_Name (string): The Provider_Name field of the event.
            EventType (string): The type of the event defined by the manifest file.
        """
        yield from self.read_etl_files(self.PATHS["boot"])



from __future__ import annotations

import re
from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

CPanelLastloginRecord = TargetRecordDescriptor(
    "application/log/cpanel/lastlogin",
    [
        ("datetime", "ts"),
        ("string", "user"),
        ("net.ipaddress", "remote_ip"),
    ],
)

CPANEL_LASTLOGIN = ".lastlogin"
CPANEL_LOGS_PATH = "/usr/local/cpanel/logs"
CPANEL_LASTLOGIN_PATTERN = re.compile(
    r"([^\s]+) # ([0-9]{4}-[0-9]{2}-[0-9]{2}) ([0-9]{2}:[0-9]{2}:[0-9]{2}) ([+-][0-9]{4})"
)


class CPanelPlugin(Plugin):
    """cPanel webhosting plugin."""

    # TODO: Parse other log files https://support.cartika.com/portal/en/kb/articles/whm-cpanel-log-files-and-locations
    __namespace__ = "cpanel"

    def check_compatible(self) -> None:
        if not self.target.fs.path(CPANEL_LOGS_PATH).exists():
            raise UnsupportedPluginError("No cPanel log path found")

    @export(record=CPanelLastloginRecord)
    def lastlogin(self) -> Iterator[CPanelLastloginRecord]:
        """Return the content of the cPanel lastlogin file.

        The lastlogin files tracks successful cPanel interface logons. New logon events are only tracked
        if the IP-address of the logon changes.

        References:
            - https://forums.cpanel.net/threads/cpanel-control-panel-last-login-clarification.579221/
            - https://forums.cpanel.net/threads/lastlogin.707557/
        """
        for user_details in self.target.user_details.all_with_home():
            if (lastlogin := user_details.home_path.joinpath(CPANEL_LASTLOGIN)).exists():
                try:
                    for index, line in enumerate(lastlogin.open("rt")):
                        line = line.strip()
                        if not line:
                            continue

                        if events := CPANEL_LASTLOGIN_PATTERN.findall(line):
                            for event in events:
                                remote_ip, date, time, utc_offset = event

                                timestamp = datetime.strptime(f"{date} {time} {utc_offset}", "%Y-%m-%d %H:%M:%S %z")

                                yield CPanelLastloginRecord(
                                    ts=timestamp,
                                    user=user_details.user.name,
                                    remote_ip=remote_ip,
                                    _target=self.target,
                                )
                        else:
                            self.target.log.warning(
                                "The cPanel lastlogin line number %s is malformed: %s", index + 1, lastlogin
                            )

                except Exception:
                    self.target.log.warning(
                        "An error occurred parsing cPanel lastlogin line number %i in file: %s", index + 1, lastlogin
                    )



from __future__ import annotations

import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.archive import vbk

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import (
    DirEntry,
    Filesystem,
    FilesystemEntry,
)
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class VbkFilesystem(Filesystem):
    """Filesystem implementation for VBK files."""

    __type__ = "vbk"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.vbk = vbk.VBK(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        try:
            vbk.VBK(fh)
        except (vbk.VBKError, EOFError):
            return False
        else:
            return True

    def get(self, path: str) -> FilesystemEntry:
        return VbkFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: vbk.DirItem | None = None) -> FilesystemEntry:
        try:
            return self.vbk.get(path, node)
        except vbk.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except vbk.IsADirectoryError as e:
            raise IsADirectoryError(path) from e
        except vbk.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except vbk.Error as e:
            raise FilesystemError(path) from e


class VbkDirEntry(DirEntry):
    fs: VbkFilesystem
    entry: vbk.DirItem

    def get(self) -> VbkFilesystemEntry:
        return VbkFilesystemEntry(self.fs, self.path, self.entry)

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        return self.entry.is_dir()

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        return self.entry.is_file()

    def is_symlink(self) -> bool:
        return False

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class VbkFilesystemEntry(FilesystemEntry):
    fs: VbkFilesystem
    entry: vbk.DirItem

    def get(self, path: str) -> FilesystemEntry:
        return VbkFilesystemEntry(
            self.fs,
            fsutil.join(self.path, path, alt_separator=self.fs.alt_separator),
            self.fs._get_node(path, self.entry),
        )

    def open(self) -> None:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self.entry.open()

    def scandir(self) -> Iterator[VbkDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry in self.entry.iterdir():
            yield VbkDirEntry(self.fs, self.path, entry.name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        return self.entry.is_dir()

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return self.entry.is_file()

    def is_symlink(self) -> bool:
        return False

    def readlink(self) -> str:
        raise NotASymlinkError

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        mode = stat.S_IFDIR if self.is_dir() else stat.S_IFREG
        size = 0 if self.is_dir() else self.entry.size

        # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
        st_info = [
            mode | 0o755,
            fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator),
            id(self.fs),
            1,
            0,
            0,
            size,
            0,
            0,
            0,
        ]

        return fsutil.stat_result(st_info)



#!/usr/bin/env python
from __future__ import annotations

import argparse
import logging
import pathlib
import sys
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from flow.record import Record, RecordPrinter, RecordStreamWriter, RecordWriter

from dissect.target.exceptions import (
    FatalError,
    PluginNotFoundError,
    TargetError,
    UnsupportedPluginError,
)
from dissect.target.helpers import cache, record_modifier
from dissect.target.helpers.logging import get_logger
from dissect.target.plugin import (
    PLUGINS,
    FunctionDescriptor,
)
from dissect.target.target import Target
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    configure_plugin_arguments,
    execute_function_on_target,
    find_and_filter_plugins,
    open_targets,
    persist_execution_report,
    process_generic_arguments,
    process_plugin_arguments,
)
from dissect.target.tools.utils.report import ExecutionReport

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record.adapter import AbstractWriter


log = get_logger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


def record_output(strings: bool = False, json: bool = False) -> AbstractWriter:
    if json:
        return RecordWriter("jsonfile://-")

    fp = sys.stdout.buffer

    if strings or fp.isatty():
        return RecordPrinter(fp)

    return RecordStreamWriter(fp)


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
        add_help=False,
    )
    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to load")
    parser.add_argument("--direct", action="store_true", help="treat TARGETS as paths to pass to plugins directly")

    configure_plugin_arguments(parser)

    parser.add_argument("-s", "--strings", action="store_true", help="print output as string")
    parser.add_argument("-d", "--delimiter", default=" ", action="store", metavar="','")
    parser.add_argument("-j", "--json", action="store_true", help="output records as json")

    parser.add_argument("--limit", type=int, help="limit number of produced records")
    parser.add_argument("--no-cache", "--ignore-cache", action="store_true", help="do not use file based caching")
    parser.add_argument(
        "--only-read-cache",
        action="store_true",
        help="only read cache files, never write them (has no effect if --no-cache is specified",
    )
    parser.add_argument(
        "--rewrite-cache",
        action="store_true",
        help=(
            "force cache files to be rewritten (has no effect if either --no-cache or --only-read-cache are specified)"
        ),
    )
    parser.add_argument("--cmdb", action="store_true")
    parser.add_argument("--hash", action="store_true", help="hash all paths in records")
    parser.add_argument("--resolve", action="store_true", help="resolve all paths in records")
    parser.add_argument(
        "--report-dir",
        type=pathlib.Path,
        help="write the query report file to the given directory",
    )

    advanced_group = parser.add_argument_group("Advanced options")
    advanced_group.add_argument(
        "--direct-sensitive", action="store_true", help="same as --direct, but paths will be case-sensitive"
    )

    configure_generic_arguments(parser)

    args, rest = parser.parse_known_args()

    # Show help for target-query
    if not args.function and ("-h" in rest or "--help" in rest):
        parser.print_help()
        return 0

    process_generic_arguments(parser, args)

    if args.no_cache:
        cache.IGNORE_CACHE = True

    if args.only_read_cache:
        cache.ONLY_READ_CACHE = True
        if args.no_cache:
            log.warning("The --only-read-cache option will be ignored as --no-cache is specified")

    if args.rewrite_cache:
        cache.REWRITE_CACHE = True
        if args.no_cache or args.only_read_cache:
            log.warning(
                "The --rewrite-cache option will be ignored as --no-cache or --only-read-cache are specified",
            )

    # Process plugin arguments after host and child args are checked
    different_output_types = process_plugin_arguments(parser, args, rest)

    if not args.targets:
        parser.error("too few arguments - missing targets")

    if args.report_dir and not args.report_dir.is_dir():
        parser.error(f"--report-dir {args.report_dir} is not a valid directory")

    default_output_type = None
    if len(different_output_types) > 1:
        log.warning("Mixed output types detected: %s, only outputting records", ",".join(different_output_types))
        default_output_type = "record"

    execution_report = ExecutionReport()
    execution_report.set_cli_args(args)
    execution_report.set_event_callbacks(Target)

    try:
        for target in open_targets(args):
            record_entries: list[tuple[FunctionDescriptor, Iterator[Record]]] = []
            basic_entries = []
            yield_entries = []

            if args.dry_run:
                print("Dry run on:", target)

            first_seen_output_type = default_output_type

            for func_def in find_and_filter_plugins(args.function, target, args.excluded_functions):
                # If the default type is record (meaning we skip everything else)
                # and actual output type is not record, continue.
                # We perform this check here because plugins that require output files/dirs
                # will exit if we attempt to exec them without (because they are implied by the wildcard).
                # Also this saves cycles of course.
                if default_output_type == "record" and func_def.output != "record":
                    continue

                if args.dry_run:
                    print(f"  execute: {func_def.name} ({func_def.path})")
                    continue

                try:
                    output_type, result = execute_function_on_target(target, func_def)
                except UnsupportedPluginError as e:
                    target.log.error(  # noqa: TRY400
                        "Unsupported plugin for %s: %s",
                        func_def.name,
                        e.root_cause_str(),
                    )
                    target.log.debug("%s", func_def, exc_info=e)
                    continue
                except PluginNotFoundError:
                    target.log.error("Cannot find plugin `%s`", func_def)  # noqa: TRY400
                    continue
                except FatalError as e:
                    e.emit_last_message(target.log.error)
                    return 1
                except Exception as e:
                    target.log.error("Exception while executing function %s (%s): %s", func_def.name, func_def.path, e)  # noqa: TRY400
                    target.log.debug("", exc_info=e)
                    target.log.debug("Function info: %s", func_def)
                    continue

                if first_seen_output_type and output_type != first_seen_output_type:
                    target.log.error(
                        (
                            "Can't mix functions that generate different outputs: output type `%s` from `%s` "
                            "does not match first seen output `%s`."
                        ),
                        output_type,
                        func_def,
                        first_seen_output_type,
                    )
                    return 0

                if not first_seen_output_type:
                    first_seen_output_type = output_type

                if output_type == "record":
                    record_entries.append((func_def, result))
                elif output_type == "yield":
                    yield_entries.append(result)
                elif output_type == "none":
                    target.log.info("No result for function `%s` (output type is set to 'none')", func_def)
                    continue
                else:
                    basic_entries.append(result)

            # Write basic functions
            if len(basic_entries) > 0:
                basic_entries_delim = args.delimiter.join(map(str, basic_entries))
                if not args.cmdb:
                    print(f"{target} {basic_entries_delim}")
                else:
                    print(f"{target.path}{args.delimiter}{basic_entries_delim}")

            # Write yield functions
            for entry in yield_entries:
                for e in entry:
                    print(e)

            # Write records
            count = 0
            break_out = False

            modifier_type = None

            if args.resolve:
                modifier_type = record_modifier.Modifier.RESOLVE

            if args.hash:
                modifier_type = record_modifier.Modifier.HASH

            modifier_func = record_modifier.get_modifier_function(modifier_type)

            if not record_entries:
                continue

            rs = record_output(args.strings, args.json)
            for func_def, record_generator in record_entries:
                try:
                    for record in record_generator:
                        rs.write(modifier_func(target, record))
                        count += 1
                        if args.limit is not None and count >= args.limit:
                            break_out = True
                            break

                except Exception as e:
                    # Ignore errors if multiple functions or multiple targets
                    if len(record_entries) > 1 or len(args.targets) > 1:
                        target.log.error(  # noqa: TRY400
                            "Exception occurred while processing output of %s.%s: %s",
                            func_def.qualname,
                            func_def.name,
                            e,
                        )
                        target.log.debug("", exc_info=e)
                    else:
                        raise

                if break_out:
                    break

    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    timestamp = datetime.now(tz=timezone.utc)

    execution_report.set_plugin_stats(PLUGINS)
    log.trace("%s", execution_report.get_formatted_report())
    if args.report_dir:
        persist_execution_report(
            args.report_dir,
            {
                "timestamp": timestamp.isoformat(),
                **execution_report.as_dict(),
            },
            timestamp=timestamp,
        )

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows.search import SearchIndexPlugin
from tests._utils import absolute_path
from tests.conftest import add_win_user

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.helpers.regutil import VirtualHive
    from dissect.target.target import Target


def test_windows_search_esedb(
    target_win: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive, hive_hku: VirtualHive
) -> None:
    """Test Windows Search EseDB parsing.

    Resources:
        - https://github.com/strozfriedberg/sidr/tree/main/tests/testdata
    """

    user_sid = "S-1-5-21-29705265-400737687-482427116-1001"

    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid=user_sid,
        home="C:\\Users\\User",
    )

    fs_win.map_file(
        "ProgramData/Microsoft/Search/Data/Applications/Windows/Windows.edb",
        str(absolute_path("_data/plugins/os/windows/search/Windows.edb")),
    )

    fs_win.map_file(
        f"Users/User/AppData/Roaming/Microsoft/Search/Data/Applications/{user_sid}/Windows.edb",
        str(absolute_path("_data/plugins/os/windows/search/Windows.edb")),
    )

    plugin = target_win.add_plugin(SearchIndexPlugin)
    records = list(target_win.search())

    # Test find_databases deduplication
    assert len(plugin.databases) == 1

    len_records = len(records)
    assert len_records == 1183 - 2  # Database contains two empty records.

    # SearchIndexActivityRecord
    assert records[-1].ts_start == datetime(2023, 2, 16, 17, 30, 35, tzinfo=timezone.utc)
    assert records[-1].ts_end == datetime(2023, 2, 16, 17, 30, 37, tzinfo=timezone.utc)
    assert records[-1].duration == 20_000_000  # equals two seconds
    assert records[-1].application_name == "PowerPoint 2016"
    assert (
        records[-1].application_id == "{6D809377-6AF0-444B-8957-A3773F02200E}\\Microsoft Office\\Office16\\POWERPNT.EXE"
    )
    assert (
        records[-1].activity_id == "ECB32AF3-1440-4086-94E3-5311F97F89C4\\{ThisPCDesktopFolder}\\This is test PPT.pptx"
    )
    assert records[-1].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.edb"

    # SearchIndexRecord (file)
    assert records[1175].ts == datetime(2023, 2, 16, 14, 36, 23, 877518, tzinfo=timezone.utc)
    assert records[1175].ts_mtime == datetime(2023, 2, 16, 14, 36, 14, 922361, tzinfo=timezone.utc)
    assert records[1175].ts_btime == datetime(2023, 2, 16, 14, 35, 23, 656454, tzinfo=timezone.utc)
    assert records[1175].ts_atime == datetime(2023, 2, 16, 14, 36, 22, 893101, tzinfo=timezone.utc)
    assert records[1175].path == "C:\\Users\\testuser\\Desktop\\StrozFriedberg-Example.txt"
    assert records[1175].type == "text/plain"
    assert records[1175].size == 50
    assert records[1175].data == "Example File from Stroz Friedberg.\r\nHappy Testing!"
    assert records[1175].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.edb"

    # SearchIndexRecord (file)
    assert records[1120].ts == datetime(2023, 2, 14, 15, 14, 40, 984222, tzinfo=timezone.utc)
    assert records[1120].ts_mtime == datetime(2023, 2, 14, 15, 3, 59, 352678, tzinfo=timezone.utc)
    assert records[1120].ts_btime == datetime(2023, 2, 14, 13, 56, 55, 827263, tzinfo=timezone.utc)
    assert records[1120].ts_atime == datetime(2023, 2, 14, 15, 12, 4, 400179, tzinfo=timezone.utc)
    assert records[1120].path == "C:\\Users\\testuser\\Desktop\\Content-Check\\Malicious.js"
    assert records[1120].type == "JavaScript File"  # no mimetype available?
    assert records[1120].size == 511
    assert (
        records[1120].data
        == "Line 1\r\nLine 2\r\nLine 3\r\nLine 4\r\nLine 5\r\nLine 6\r\nLine 7\r\nLine 8\r\nLine 9\r\nLine 10\r\nLine 11\r\nLine 12\r\nLine 13\r\nLine 14\r\nLine 15\r\nLine 16\r\nLine 17\r\nLine 18\r\nLine 19\r\nLine 20\r\nLine 21\r\nLine 22\r\nLine 23\r\nLine 24\r\nLine 25\r\nLine 26\r\nLine 27\r\nLine 28\r\nLine 29\r\nLine 30\r\nLine 31\r\nLine 32\r\nLine 33\r\nLine 34\r\nLine 35\r\nLine 36\r\nLine 37\r\nLine 38\r\nLine 39\r\nLine 40\r\nLine 41\r\nLine 42\r\nLine 43\r\nLine 44\r\nLine 45\r\nLine 46\r\nLine 47\r\nLine 48\r\nLine 49\r\nLine 50\r\nLine 51\r\nLine 52\r\nLine 53\r\nLine 54\r\nLine 55\r\nLine 56\r\nLine 57\r\nLine \r\n"  # noqa: E501
    )
    assert records[1120].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.edb"

    # SearchIndexRecord (folder)
    assert records[1116].ts == datetime(2023, 2, 14, 13, 59, 31, 356632, tzinfo=timezone.utc)
    assert records[1116].ts_mtime == datetime(2023, 2, 14, 13, 57, 29, 517132, tzinfo=timezone.utc)
    assert records[1116].ts_btime == datetime(2023, 2, 14, 13, 57, 8, 689777, tzinfo=timezone.utc)
    assert records[1116].ts_atime == datetime(2023, 2, 14, 13, 59, 31, 340546, tzinfo=timezone.utc)
    assert records[1116].path == "C:\\Users\\testuser\\Desktop\\Content-Check"
    assert records[1116].type == "File folder"  # no mimetype available?
    assert not records[1116].size
    assert not records[1116].data
    assert records[1116].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.edb"

    # BrowserHistoryRecord (iehistory)
    assert records[995].ts == datetime(2023, 2, 13, 14, 17, 22, 448000, tzinfo=timezone.utc)
    assert records[995].browser == "iehistory"
    assert (
        records[995].url
        == "https://support.microsoft.com/en-us/microsoft-edge/this-website-doesn-t-work-in-internet-explorer-8f5fc675-cd47-414c-9535-12821ddfc554?ui=en-us&rs=en-us&ad=us"
    )
    assert records[995].host == "support.microsoft.com"
    assert records[995].title == "This website doesn't work in Internet Explorer - Microsoft Support"
    assert records[995].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.edb"
    assert records[995].user_id == user_sid


def test_windows_search_sqlite(
    target_win: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive, hive_hku: VirtualHive
) -> None:
    """Test Windows 11 Search SQLite3 parsing.

    Resources:
        - https://github.com/strozfriedberg/sidr/tree/main/tests/testdata
    """

    user_sid = "S-1-5-21-4268361623-692440835-3372367631-1001"

    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid=user_sid,
        home="C:\\Users\\User",
    )

    fs_win.map_file(
        "ProgramData/Microsoft/Search/Data/Applications/Windows/Windows.db",
        str(absolute_path("_data/plugins/os/windows/search/Windows.db")),
    )

    target_win.add_plugin(SearchIndexPlugin)
    records = list(target_win.search())

    len_records = len(records)
    assert len_records == 839 - 1  # Database contains one empty record.

    # SearchIndexActivityRecord
    assert records[698].ts_start == datetime(2023, 1, 30, 22, 14, 18, tzinfo=timezone.utc)
    assert records[698].ts_end == datetime(2023, 1, 30, 22, 14, 20, tzinfo=timezone.utc)
    assert records[698].duration == 20_000_000
    assert records[698].application_name == "notepad++.exe"
    assert records[698].application_id == "{6D809377-6AF0-444B-8957-A3773F02200E}\\Notepad++\\notepad++.exe"
    assert records[698].activity_id == "ECB32AF3-1440-4086-94E3-5311F97F89C4\\{Public}\\Threat\\becon.xml"
    assert records[698].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.db"

    # SearchIndexRecord (file)
    assert records[837].ts == datetime(2023, 1, 31, 2, 45, 2, 871614, tzinfo=timezone.utc)
    assert records[837].ts_mtime == datetime(2023, 1, 31, 2, 45, 2, 56444, tzinfo=timezone.utc)
    assert records[837].ts_btime == datetime(2023, 1, 31, 2, 26, 28, 898306, tzinfo=timezone.utc)
    assert records[837].ts_atime == datetime(2023, 1, 31, 2, 45, 2, 56444, tzinfo=timezone.utc)
    assert records[837].path == "C:\\Users\\Public\\malware\\New-beacon.xml"
    assert records[837].type == "text/xml"
    assert int(records[837].size) == 174
    assert not records[837].data
    assert records[837].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.db"

    # BrowserHistoryRecord (edge)
    assert records[711].ts == datetime(2023, 1, 31, 0, 9, 47, 972897, tzinfo=timezone.utc)
    assert records[711].browser == "winrt"
    assert (
        records[711].url
        == "https://www.bing.com/search?q=install+chrome&cvid=2ce0f71581824fda82398075bb250924&aqs=edge.0.0j69i57j0l7.2774j0j7&FORM=ANNTA0&PC=U531"
    )
    assert records[711].host == "www.bing.com"
    assert records[711].source == "\\sysvol\\ProgramData\\Microsoft\\Search\\Data\\Applications\\Windows\\Windows.db"
    assert records[711].user_id == user_sid



from __future__ import annotations

import os
import textwrap
from typing import TYPE_CHECKING

from dissect.target.helpers.cyber import cyber
from dissect.target.loader import Loader
from dissect.target.loader import open as loader_open
from dissect.target.loaders.raw import RawLoader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target

HEADER = r"""
┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
┃      _______     ______  ______ _____      ┃
┃     / ____\ \   / /  _ \|  ____|  __ \     ┃
┃    | |     \ \_/ /| |_) | |__  | |__) |    ┃
┃    | |      \   / |  _ <|  __| |  _  /     ┃
┃    | |____   | |  | |_) | |____| | \ \     ┃
┃     \_____|  |_|  |____/|______|_|  \_\    ┃
┃                                            ┃
┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛

"""


class CyberLoader(Loader):
    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self._real = loader_open(path) or RawLoader(path)

    @staticmethod
    def detect(path: Path) -> bool:
        return False

    def map(self, target: Target) -> None:
        try:
            cols, _ = os.get_terminal_size()
            width = HEADER.index("\n", 1)
            header = textwrap.indent(HEADER, " " * ((cols - width) // 2))
        except OSError:
            header = HEADER

        with cyber(mask_space=True, mask_indent=False):
            print(header)

        target.props["cyber"] = True
        return self._real.map(target)
